/* $Id: pftop.c,v 1.23 2002/09/02 20:30:22 canacar Exp $	 */
/*
 * Copyright (c) 2001 Can Erkin Acar
 * Copyright (c) 2001 Daniel Hartmeier
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <net/pfvar.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <curses.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <err.h>
#include <signal.h>
#include <ctype.h>

#define MIN_NUM_STATES 1024
#define NUM_STATE_INC  1024

#define HEADER_LINES	4
#define HOME_X	0
#define HOME_Y	2

#define CTRL_A  1
#define CTRL_E  5
#define CTRL_L  12
#define CTRL_N  14
#define CTRL_P  16
#define CTRL_V  22

#define META_V  246

/* view management */
int read_states(int dev);
void sort_states(void);
void print_states(void);
int read_rules(int dev);
void sort_rules(void);
void print_rules(void);

/* qsort callbacks */
int qsort_size_callback(const void *s1, const void *s2);
int qsort_exp_callback(const void *s1, const void *s2);
int qsort_pkt_callback(const void *s1, const void *s2);
int qsort_age_callback(const void *s1, const void *s2);
int qsort_sa_callback(const void *s1, const void *s2);
int qsort_sp_callback(const void *s1, const void *s2);
int qsort_da_callback(const void *s1, const void *s2);
int qsort_dp_callback(const void *s1, const void *s2);


struct pf_state *state_buf = NULL;
int state_buf_len = 0;
u_int32_t num_states = 0;
u_int32_t num_disp = 0;
u_int32_t num_rules = 0;
int columns, lines;
int interactive = 1;
int sortdir = 1;
int rawmode = 0;
int maxprint = 0;
int dispstart = 0;

volatile sig_atomic_t gotsig_close = 0;
volatile sig_atomic_t gotsig_resize = 0;
volatile sig_atomic_t gotsig_alarm = 0;
int need_update = 0;
int need_sort = 0;

#define FLD_ALIGN_LEFT   0
#define FLD_ALIGN_RIGHT  1
#define FLD_ALIGN_CENTER 2

SCREEN *screen;
typedef struct {
	char *title;
	int norm_width;
	int max_width;
	int align;
	int start;
	int width;
} field_def;
/* for states */
#define FLD_SRC     0
#define FLD_DEST    1
#define FLD_GW      2
#define FLD_STATE   3
#define FLD_AGE     4
#define FLD_EXP     5
/* common */
#define FLD_PROTO   6
#define FLD_DIR     7
#define FLD_PKTS    8
#define FLD_BYTES   9
#define FLD_RULE   10
/* for rules */
#define FLD_LABEL  11
#define FLD_STATS  12
#define FLD_EVAL   13
#define FLD_ACTION 14
#define FLD_LOG    15
#define FLD_QUICK  16
#define FLD_KST    17

#if OS_LEVEL < 31
#define PF_RULE_LABEL_SIZE 20
#endif

field_def fields[] = {
	{"SRC", 20, 45, FLD_ALIGN_LEFT, -1, 0},
	{"DEST", 20, 45, FLD_ALIGN_LEFT, -1, 0},
	{"GW", 20, 45, FLD_ALIGN_LEFT, -1, 0},
	{"STATE", 5, 5, FLD_ALIGN_CENTER, -1, 0},
	{"AGE", 5, 8, FLD_ALIGN_RIGHT, -1, 0},
	{"EXP", 5, 8, FLD_ALIGN_RIGHT, -1, 0},
	{"PR ", 4, 4, FLD_ALIGN_RIGHT, -1, 0},
	{"DIR", 1, 3, FLD_ALIGN_CENTER, -1, 0},
	{"PKTS", 5, 8, FLD_ALIGN_RIGHT, -1, 0},
	{"BYTES", 5, 8, FLD_ALIGN_RIGHT, -1, 0},
	{"RULE", 2, 4, FLD_ALIGN_RIGHT, -1, 0},
	{"LABEL", 20, PF_RULE_LABEL_SIZE, FLD_ALIGN_LEFT, -1, 0},
	{"STATES", 5, 8, FLD_ALIGN_RIGHT, -1, 0},
	{"EVAL", 5, 8, FLD_ALIGN_RIGHT, -1, 0},
	{"ACT", 1, 5, FLD_ALIGN_LEFT, -1, 0},
	{"LOG", 1, 3, FLD_ALIGN_LEFT, -1, 0},
	{"QUICK", 1, 1, FLD_ALIGN_LEFT, -1, 0},
	{"KS", 1, 1, FLD_ALIGN_LEFT, -1, 0}
};

int *fieldseq = 0;

int view0[] = {
	FLD_PROTO, FLD_DIR, FLD_SRC, FLD_DEST, FLD_STATE,
	FLD_AGE, FLD_EXP, FLD_PKTS, FLD_BYTES, -1
};
int view1[] = {
	FLD_PROTO, FLD_DIR, FLD_SRC, FLD_DEST, FLD_GW, FLD_STATE,
	FLD_AGE, FLD_EXP, FLD_PKTS, FLD_BYTES, FLD_RULE, -1
};
int view2[] = {
	FLD_PROTO, FLD_DIR, FLD_SRC, FLD_DEST, FLD_STATE,
	FLD_AGE, FLD_EXP, FLD_PKTS, FLD_BYTES, FLD_RULE, FLD_GW, -1
};
int view3[] = {
	FLD_PROTO, FLD_DIR, FLD_SRC, FLD_DEST, FLD_AGE, FLD_EXP,
	FLD_PKTS, FLD_BYTES, FLD_STATE, FLD_RULE, FLD_GW, -1
};
int view4[] = {
	FLD_PROTO, FLD_DIR, FLD_SRC, FLD_DEST, FLD_PKTS, FLD_BYTES,
	FLD_STATE, FLD_AGE, FLD_EXP, FLD_RULE, FLD_GW, -1
};

int view5[] = {
	FLD_RULE, FLD_ACTION, FLD_DIR, FLD_LOG, FLD_QUICK, FLD_PROTO,
	FLD_KST, FLD_PKTS, FLD_BYTES, FLD_STATS, FLD_LABEL, -1
};

int view6[] = {
	FLD_RULE, FLD_LABEL, FLD_PKTS, FLD_BYTES, FLD_STATS, 
	FLD_ACTION, FLD_DIR, FLD_LOG, FLD_QUICK, FLD_PROTO,
	FLD_KST,  -1
};

struct view_manager {
	char *name;
	int  (*read_fn) (int);
	void (*sort_fn) (void);
	void (*print_fn) (void);
};

struct view_manager state_mgr = {
	"State", read_states, sort_states, print_states
};

struct view_manager rule_mgr = {
	"Rule", read_rules, sort_rules, print_rules
};

typedef struct {
	int *view;
	char *name;
	int hotkey;
	struct view_manager *mgr;
} field_view;

field_view views[] = {
	{view0, "Default", '0', &state_mgr},
	{view1, "Long", '1', &state_mgr},
	{view2, "State", '2', &state_mgr},
	{view3, "Time", '3', &state_mgr},
	{view4, "Size", '4', &state_mgr},
	{view5, "Rules", '5', &rule_mgr},
	{view6, "Label", '6', &rule_mgr},
	{NULL, NULL, 0}
};

field_view *curr_view = NULL;
struct view_manager *curr_mgr = NULL;

int num_fields = sizeof(fields) / sizeof(field_def);
int curr_line = 0;

/* line buffer for raw mode */
#define MAX_LINE_BUF 1024
char linebuf[MAX_LINE_BUF];
int linepos = 0;

/* temp storage for state printing */
char tmp_buf[MAX_LINE_BUF];

/* ordering */
typedef struct {
	char *name;
	char *match;
	int hotkey;
	int (*func) (const void *, const void *);
} order_type;

order_type order_list[] = {
	{"None", "none", 'n', NULL},
	{"Bytes", "bytes", 'b', qsort_size_callback},
	{"Expiry", "exp", 'e', qsort_exp_callback},
	{"Packets", "pkt", 'p', qsort_pkt_callback},
	{"Age", "age", 'a', qsort_age_callback},
	{"Source addr", "src", 's', qsort_sa_callback},
	{"Dest. addr", "dest", 'd', qsort_da_callback}, 
	{"Source port", "sport", 'S', qsort_sp_callback},
	{"Dest. port", "dport", 'D', qsort_dp_callback}, 
	{0, 0, 0}
};

order_type *ordering = NULL;


/* screen output functions */

void
move_horiz(int offset)
{
	if (rawmode) {
		if (offset <= 0)
			linepos = 0;
		else if (offset >= MAX_LINE_BUF)
			linepos = MAX_LINE_BUF - 1;
		else
			linepos = offset;
	} else {
		move(curr_line, offset);
	}
}

void
print_str(int len, char *str)
{
	int n;

	if (len <= 0)
		return;

	if (rawmode) {
		for (n = 0; n < len && linepos < MAX_LINE_BUF; n++)
			linebuf[linepos++] = str[n];
	} else {
		for (n = 0; n < len; n++)
			addch((chtype) str[n]);
	}
}

inline void
clear_linebuf(void)
{
	memset(linebuf, ' ', MAX_LINE_BUF);
}

void
end_line(void)
{
	if (rawmode) {
		if (linepos >= MAX_LINE_BUF)
			linepos = MAX_LINE_BUF - 1;
		linebuf[linepos] = 0;
		printf("%s\n", linebuf);
		clear_linebuf();
	}
	curr_line++;
}

void
end_page(void)
{
	if (rawmode) {
		linepos = 0;
		clear_linebuf();
	} else {
		move(HOME_Y, HOME_X);
		refresh();
	}
	curr_line = 0;
}

void 
rawaddstr(char *s)
{
	if (rawmode)
		printf("%s", s);
	else
		addstr(s);
}

/* field output functions */

void
print_fld_str(int fid, char *str)
{
	int len;
	field_def *fld;

	if (str == NULL)
		return;
	if (fid < 0 || fid >= num_fields)
		return;

	fld = &fields[fid];
	if (fld->start < 0)
		return;

	len = strlen(str);
	if (len >= fld->width) {
		move_horiz(fld->start);
		print_str(fld->width, str);
	} else {
		switch (fld->align) {
		case FLD_ALIGN_RIGHT:
			move_horiz(fld->start + (fld->width - len));
			break;
		case FLD_ALIGN_CENTER:
			move_horiz(fld->start + (fld->width - len) / 2);
			break;
		default:
			move_horiz(fld->start);
			break;
		}
		print_str(len, str);
	}
};

void
print_title(void)
{
	int n;
	for (n = 0; n < num_fields; n++)
		print_fld_str(n, fields[n].title);
	end_line();
}



/* view related functions */
void
field_setup(void)
{
	int *v, n, st, fwid,  change;
	int width = columns;

	for (n = 0; n < num_fields; n++) {
		fields[n].start = -1;
		fields[n].width = fields[n].norm_width;
	}

	st = 0;
	if (curr_view == 0)
		return;

	dispstart = 0;

	for (v = curr_view->view; *v >= 0 && *v < num_fields; v++) {
		n = *v;
		fields[n].start = 1;
		fwid = fields[n].width;
		if (fwid > width) {
			if (width <= 0) {
				fields[n].start = -1;
			} else {
				fields[n].width = width;
				width = 0;
			}
			break;
		}
		fwid++;
		st += fwid;
		width -= fwid;
	}

	change = 0;
	while (width > 0) {
		change = 0;
		for (v = curr_view->view; *v >= 0 && *v < num_fields; v++) {
			n = *v;
			if (fields[n].width < fields[n].max_width) {
				fields[n].width++;
				width--;
				change=1;
			}
			if (width <= 0) break;
		}
		if (change == 0) break;
	}

	st = 0;
	for (v = curr_view->view; *v >= 0 && *v < num_fields; v++) {
		n = *v;
		if (fields[n].start < 0) break;
		fields[n].start = st;
		st+= fields[n].width+1;
	}
}

void
set_curr_view(field_view *v)
{
	if (v == NULL) {
		curr_view = NULL;
		curr_mgr = NULL;
		return;
	}

	if (curr_view != NULL)
		if (curr_view->mgr != v->mgr)
			gotsig_alarm = 1;

	curr_view = v;
	curr_mgr = v->mgr;
	field_setup();
	need_update = 1;
}

void
set_view(char *opt)
{
	field_view *v;
	set_curr_view(views);

	if (opt == NULL)
		return;

	for (v = views; v->name != NULL; v++) {
		if (strcasecmp(opt, v->name) == 0) {
			set_curr_view(v);
			return;
		}
	}
}

int
set_view_hotkey(int ch)
{
	field_view *v;
	int key = tolower(ch);
	for (v = views; v->name != NULL; v++) {
		if (key == v->hotkey) {
			set_curr_view(v);
			return 1;
		}
	}
	return 0;
}

void
next_view(void)
{
	field_view *v;
	for (v = views; v->name != NULL; v++) {
		if (curr_view == v) {
			v++;
			if (v->name == NULL)
				break;
			set_curr_view(v);
			return;
		}
	}
	set_curr_view(views);
}


/* ordering functions */
int
qsort_size_callback(const void *s1, const void *s2)
{
	if (((struct pf_state *) s2)->bytes >
	    ((struct pf_state *) s1)->bytes)
		return sortdir;
	return -sortdir;
}

int
qsort_pkt_callback(const void *s1, const void *s2)
{
	if (((struct pf_state *) s2)->packets >
	    ((struct pf_state *) s1)->packets)
		return sortdir;
	return -sortdir;
}

int
qsort_age_callback(const void *s1, const void *s2)
{
	if (((struct pf_state *) s2)->creation >
	    ((struct pf_state *) s1)->creation)
		return sortdir;
	return -sortdir;
}

int
qsort_exp_callback(const void *s1, const void *s2)
{
	if (((struct pf_state *) s2)->expire >
	    ((struct pf_state *) s1)->expire)
		return sortdir;
	return -sortdir;
}

int
compare_addr(int af, struct pf_addr *a, struct pf_addr *b)
{
	switch (af) {
	case AF_INET:
		if (ntohl(a->addr32[0]) > ntohl(b->addr32[0]))
			return 1;
		if (a->addr32[0] != b->addr32[0])
			return -1;
		break;
	case AF_INET6:
		if (ntohl(a->addr32[0]) > ntohl(b->addr32[0]))
			return 1;
		if (a->addr32[0] != b->addr32[0])
			return -1;
		if (ntohl(a->addr32[1]) > ntohl(b->addr32[1]))
			return 1;
		if (a->addr32[1] != b->addr32[1])
			return -1;
		if (ntohl(a->addr32[2]) > ntohl(b->addr32[2]))
			return 1;
		if (a->addr32[2] != b->addr32[2])
			return -1;
		if (ntohl(a->addr32[3]) > ntohl(b->addr32[3]))
			return 1;
		if (a->addr32[3] != b->addr32[3])
			return -1;
		break;
	}
	
	return 0;
}

int
qsort_sa_callback(const void *s1, const void *s2)
{
	struct pf_state_host *a, *b;
	int af, ret;

	af = ((struct pf_state *) s1)->af;

	if (af > ((struct pf_state *) s2)->af)
		return sortdir;
	if (af < ((struct pf_state *) s2)->af)
		return -sortdir;
	
	if (((struct pf_state *) s1)->direction == PF_OUT) {
                a = &((struct pf_state *)s1)->lan;
        } else {
                a = &((struct pf_state *)s1)->ext;
	}

	if (((struct pf_state *) s2)->direction == PF_OUT) {
                b = &((struct pf_state *)s2)->lan;
        } else {
                b = &((struct pf_state *)s2)->ext;
        }

	ret = compare_addr(af, &a->addr, &b->addr);
	if (ret)
		return ret * sortdir;

	if (ntohs(a->port) > ntohs(b->port))
		return sortdir;
	return -sortdir;
}

int
qsort_sp_callback(const void *s1, const void *s2)
{
	struct pf_state_host *a, *b;
	int af;

	af = ((struct pf_state *) s1)->af;

	if (af > ((struct pf_state *) s2)->af)
		return sortdir;
	if (af < ((struct pf_state *) s2)->af)
		return -sortdir;
	
	if (((struct pf_state *) s1)->direction == PF_OUT) {
                a = &((struct pf_state *)s1)->lan;
        } else {
                a = &((struct pf_state *)s1)->ext;
	}

	if (((struct pf_state *) s2)->direction == PF_OUT) {
                b = &((struct pf_state *)s2)->lan;
        } else {
                b = &((struct pf_state *)s2)->ext;
        }

	if (ntohs(a->port) > ntohs(b->port))
		return sortdir;
	if (ntohs(a->port) < ntohs(b->port))
		return -sortdir;

	if (compare_addr(af, &a->addr, &b->addr) > 0)
		return sortdir;
	return -sortdir;
}

int
qsort_da_callback(const void *s1, const void *s2)
{
	struct pf_state_host *a, *b;
	int af, ret;

	af = ((struct pf_state *) s1)->af;

	if (af > ((struct pf_state *) s2)->af)
		return sortdir;
	if (af < ((struct pf_state *) s2)->af)
		return -sortdir;
	
	if (((struct pf_state *) s1)->direction == PF_OUT) {
                a = &((struct pf_state *)s1)->ext;
        } else {
                a = &((struct pf_state *)s1)->lan;
	}

	if (((struct pf_state *) s2)->direction == PF_OUT) {
                b = &((struct pf_state *)s2)->ext;
        } else {
                b = &((struct pf_state *)s2)->lan;
        }

	ret = compare_addr(af, &a->addr, &b->addr);
	if (ret)
		return ret * sortdir;

	if (ntohs(a->port) > ntohs(b->port))
		return sortdir;
	return -sortdir;
}

int
qsort_dp_callback(const void *s1, const void *s2)
{
	struct pf_state_host *a, *b;
	int af;

	af = ((struct pf_state *) s1)->af;

	if (af > ((struct pf_state *) s2)->af)
		return sortdir;
	if (af < ((struct pf_state *) s2)->af)
		return -sortdir;
	
	if (((struct pf_state *) s1)->direction == PF_OUT) {
                a = &((struct pf_state *)s1)->ext;
        } else {
                a = &((struct pf_state *)s1)->lan;
	}

	if (((struct pf_state *) s2)->direction == PF_OUT) {
                b = &((struct pf_state *)s2)->ext;
        } else {
                b = &((struct pf_state *)s2)->lan;
        }

	if (ntohs(a->port) > ntohs(b->port))
		return sortdir;
	if (ntohs(a->port) < ntohs(b->port))
		return -sortdir;

	if (compare_addr(af, &a->addr, &b->addr) > 0)
		return sortdir;
	return -sortdir;

}

void
set_order(char *opt)
{
	order_type *o;
	ordering = order_list;

	if (opt == NULL)
		return;

	for (o = order_list; o->name != NULL; o++) {
		if (strcasecmp(opt, o->match) == 0) {
			ordering = o;
			return;
		}
	}
}

int
set_order_hotkey(int ch)
{
	order_type *o;
	int key = ch;
	for (o = order_list; o->name != NULL; o++) {
		if (key == o->hotkey) {
			if (ordering == o) {
				sortdir *= -1;
			} else {
				ordering = o;
			}
			return 1;
		}
	}
	return 0;
}

void
next_order(void)
{
	order_type *o;
	for (o = order_list; o->name != NULL; o++) {
		if (ordering == o) {
			o++;
			if (o->name == NULL)
				break;
			ordering = o;
			return;
		}
	}
	ordering = order_list;
}

void
sort_states(void)
{
	if (ordering == NULL)
		return;
	if (ordering->func == NULL)
		return;
	if (state_buf == NULL)
		return;
	if (num_states <= 0)
		return;

	qsort(state_buf, num_states, sizeof(struct pf_state), ordering->func);
}

/* state management functions */

void
alloc_buf(int ns)
{
	int len;

	if (ns < MIN_NUM_STATES)
		ns = MIN_NUM_STATES;

	len = ns * sizeof(struct pf_state);

	if (len >= state_buf_len) {
		len += NUM_STATE_INC * sizeof(struct pf_state);
		state_buf = realloc(state_buf, len);
		if (state_buf == NULL)
			err(1, "realloc");
		state_buf_len = len;
	}
}

int
read_states(int dev)
{
	struct pfioc_states ps;

	while (1) {
		ps.ps_len = state_buf_len;
		ps.ps_buf = (char *) state_buf;

		if (ioctl(dev, DIOCGETSTATES, &ps) < 0) {
			warnx("DIOCGETSTATES");
			return (-1);
		}
		num_states = ps.ps_len / sizeof(struct pf_state);

		if (ps.ps_len < state_buf_len)
			break;

		alloc_buf(num_states);
	}
	num_disp = num_states;
	return 0;
}

int
unmask(struct pf_addr * m, u_int8_t af)
{
	int i = 31, j = 0, b = 0, msize;
	u_int32_t tmp;

	if (af == AF_INET)
		msize = 1;
	else
		msize = 4;
	while (j < msize && m->addr32[j] == 0xffffffff) {
		b += 32;
		j++;
	}
	if (j < msize) {
		tmp = ntohl(m->addr32[j]);
		for (i = 31; tmp & (1 << i); --i)
			b++;
	}
	return (b);
}

/* display functions */

int
print_header(struct pf_status *status)
{
	int start = dispstart + 1;
	int end = dispstart + maxprint;

	if (end > num_disp)
		end = num_disp;

	if (rawmode) {
		printf("\n\npfTop: %s: ", curr_mgr ? curr_mgr->name : "???");
		if (num_disp == 0) 
			printf("No entries\n");
		else
			printf("%d - %d of %u \n", start, end, num_disp);
	} else {
		mvprintw(0, 0, "pfTop: %s: ",
			 curr_mgr ? curr_mgr->name : "???");
		if (num_disp == 0)
			printw("No entries\n");
		else
			printw("%u - %u of %u\n", start, end, num_disp);
	}

	rawaddstr("Order: ");

	if (ordering) {
		rawaddstr(ordering->name);
		if (sortdir < 0)
			rawaddstr(" (rev)");
	} else
		rawaddstr("None");
	if (curr_view) {
		rawaddstr("  View: ");
		rawaddstr(curr_view->name);
	}
	if (rawmode) {
		printf("\n\n");
	}
	curr_line = 3;
	print_title();

	return (0);
}

void
print_addr(struct pf_addr * addr, struct pf_addr * mask, int af)
{
	static char buf[48];
	const char *bf;

	bf = inet_ntop(af, addr, buf, sizeof(buf));
	strlcpy(tmp_buf, bf, sizeof(tmp_buf));

	if (mask != NULL) {
		if (!PF_AZERO(mask, af))
			snprintf(buf, sizeof(buf), "/%u", unmask(mask, af));
		strlcat(tmp_buf, buf, sizeof(tmp_buf));
	}
}

void
print_fld_host(int fid, struct pf_state_host * h, int af)
{
	int sz, len;

	u_int16_t p = ntohs(h->port);

	if (fid < 0 || fid >= num_fields)
		return;

	if (fields[fid].width < 3) {
		print_fld_str(fid, "*");
		return;
	}
	print_addr(&h->addr, NULL, af);

	sz = strlen(tmp_buf);
	len = sizeof(tmp_buf) - sz;

	if (af == AF_INET)
		snprintf(tmp_buf + sz, len, ":%u", p);
	else
		snprintf(tmp_buf + sz, len, "[%u]", p);

	print_fld_str(fid, tmp_buf);
}

void
print_fld_age(int fid, unsigned age)
{
	int len;

	if (fid < 0 || fid >= num_fields)
		return;
	len = fields[fid].width;
	if (len < 1)
		return;

	snprintf(tmp_buf, sizeof(tmp_buf), "%u", age);
	if (strlen(tmp_buf) <= len)
		goto ok;

	age /= 60;
	snprintf(tmp_buf, sizeof(tmp_buf), "%um", age);
	if (strlen(tmp_buf) <= len)
		goto ok;
	if (age == 0)
		goto err;

	age /= 60;
	snprintf(tmp_buf, sizeof(tmp_buf), "%uh", age);
	if (strlen(tmp_buf) <= len)
		goto ok;
	if (age == 0)
		goto err;

	age /= 24;
	snprintf(tmp_buf, sizeof(tmp_buf), "%ud", age);
	if (strlen(tmp_buf) <= len)
		goto ok;

err:
	print_fld_str(fid, "*");
	return;

ok:
	print_fld_str(fid, tmp_buf);
}

void
print_fld_size(int fid, unsigned size)
{
	int len;

	if (fid < 0 || fid >= num_fields)
		return;
	len = fields[fid].width;
	if (len < 1)
		return;

	snprintf(tmp_buf, sizeof(tmp_buf), "%u", size);
	if (strlen(tmp_buf) <= len)
		goto ok;

	size /= 1024;
	snprintf(tmp_buf, sizeof(tmp_buf), "%uK", size);
	if (strlen(tmp_buf) <= len)
		goto ok;
	if (size == 0)
		goto err;

	size /= 1024;
	snprintf(tmp_buf, sizeof(tmp_buf), "%uM", size);
	if (strlen(tmp_buf) <= len)
		goto ok;
	if (size == 0)
		goto err;

	size /= 1024;
	snprintf(tmp_buf, sizeof(tmp_buf), "%uG", size);
	if (strlen(tmp_buf) <= len)
		goto ok;

err:
	print_fld_str(fid, "*");
	return;

ok:
	print_fld_str(fid, tmp_buf);
}

void
print_fld_uint(int fid, unsigned size)
{
	int len;

	if (fid < 0 || fid >= num_fields)
		return;

	len = fields[fid].width;
	if (len < 1)
		return;

	snprintf(tmp_buf, sizeof(tmp_buf), "%u", size);

	if (strlen(tmp_buf) > len)
		print_fld_str(fid, "*");
	else
		print_fld_str(fid, tmp_buf);
}

int
print_state(struct pf_state * s)
{
	struct pf_state_peer *src, *dst;

	if (s->direction == PF_OUT) {
		src = &s->src;
		dst = &s->dst;
	} else {
		src = &s->dst;
		dst = &s->src;
	}
	switch (s->proto) {
	case IPPROTO_TCP:
		print_fld_str(FLD_PROTO, "TCP");
		break;
	case IPPROTO_UDP:
		print_fld_str(FLD_PROTO, "UDP");
		break;
	case IPPROTO_ICMPV6:
	case IPPROTO_ICMP:
		print_fld_str(FLD_PROTO, "ICMP");
		break;
	default:
		print_fld_str(FLD_PROTO, "????");
		break;
	}

	if (s->direction == PF_OUT) {
		print_fld_host(FLD_SRC, &s->lan, s->af);
		print_fld_host(FLD_DEST, &s->ext, s->af);
	} else {
		print_fld_host(FLD_SRC, &s->ext, s->af);
		print_fld_host(FLD_DEST, &s->lan, s->af);
	}

	if (PF_ANEQ(&s->lan.addr, &s->gwy.addr, s->af) ||
	    (s->lan.port != s->gwy.port)) {
		print_fld_host(FLD_GW, &s->gwy, s->af);
	}

	if (s->direction == PF_OUT)
		print_fld_str(FLD_DIR, "Out");
	else
		print_fld_str(FLD_DIR, "In");

	snprintf(tmp_buf, MAX_LINE_BUF, "%u:%u", src->state, dst->state);
	print_fld_str(FLD_STATE, tmp_buf);

	print_fld_age(FLD_AGE, s->creation);
	print_fld_age(FLD_EXP, s->expire);
	print_fld_size(FLD_PKTS, s->packets);
	print_fld_size(FLD_BYTES, s->bytes);
#if OS_LEVEL > 31
	print_fld_uint(FLD_RULE, s->rule.nr);
#endif
	end_line();
	return 1;
}

void
print_states(void)
{
	int n, count = 0;

	for (n = dispstart; n < num_disp; n++) {
		count += print_state(state_buf + n);
		if (maxprint > 0 && count >= maxprint)
			break;
	}
}

int
disp_update(int dev)
{
	struct pf_status status;

	if (dispstart + maxprint > num_disp)
		dispstart = num_disp - maxprint;
	
	if (dispstart < 0)
		dispstart = 0;

	if (ioctl(dev, DIOCGETSTATUS, &status)) {
		warnx("DIOCGETSTATUS");
		return (-1);
	}

	print_header(&status);

	if (curr_view == NULL)
		return 0;

	if (curr_mgr != NULL)
		if (curr_mgr->print_fn != NULL)
		curr_mgr->print_fn();

	return (0);
}

/* rule display */

struct pf_rule *rules = NULL;
u_int32_t alloc_rules = 0;

int
read_rules(int dev)
{
	struct pfioc_rule pr;
	u_int32_t nr;

	if (ioctl(dev, DIOCGETRULES, &pr)) {
		warnx("DIOCGETRULES");
		return (-1);
	}
	num_rules = pr.nr;

	if (num_rules < 0) return (0);

	if (rules == NULL) {
		rules = malloc(num_rules * sizeof(struct pf_rule));
		if (rules == NULL)
			err(1, "malloc");
		alloc_rules = num_rules;
	} else if (num_rules > num_rules) {
		rules = realloc(rules, num_rules * sizeof(struct pf_rule));
		if (rules == NULL)
			err(1, "realloc");
		alloc_rules = num_rules;
	}

	for (nr = 0; nr < num_rules; ++nr) {
		pr.nr = nr;
		if (ioctl(dev, DIOCGETRULE, &pr)) {
			warnx("DIOCGETRULE");
			return (-1);
		}
		rules[nr] = pr.rule;
	}
	num_disp = num_rules;
	return (0);
}

void
sort_rules(void)
{

}

void
print_rule(struct pf_rule *pr)
{
	if (pr == NULL) return;

#if OS_LEVEL > 30
	print_fld_str(FLD_LABEL, pr->label);
#endif
#if OS_LEVEL > 31
	print_fld_size(FLD_STATS, pr->states);
#endif
	print_fld_size(FLD_PKTS, pr->packets);
	print_fld_size(FLD_BYTES, pr->bytes);
	print_fld_uint(FLD_RULE, pr->nr);
	print_fld_str(FLD_DIR, pr->direction ? "Out" : "In");
	if (pr->quick)
		print_fld_str(FLD_QUICK, "Quick");

	if (pr->keep_state == PF_STATE_NORMAL)
		print_fld_str(FLD_KST, "Keep");
	else if (pr->keep_state == PF_STATE_MODULATE)
		print_fld_str(FLD_KST, "Mod");
	
	if (pr->log == 1)
		print_fld_str(FLD_LOG, "Log");
	else if (pr->log == 2)
		print_fld_str(FLD_LOG, "All");

	if (pr->action == PF_PASS)
		print_fld_str(FLD_ACTION, "Pass");
	else if (pr->action == PF_DROP)
		print_fld_str(FLD_ACTION, "Block");
	else print_fld_str(FLD_ACTION, "Scrub");


	if (pr->proto) {
		struct protoent *p = getprotobynumber(pr->proto);

		if (p != NULL)
			print_fld_str(FLD_PROTO, p->p_name);
		else
			print_fld_uint(FLD_PROTO, pr->proto);
	}

	end_line();
}

void
print_rules(void)
{
	u_int32_t n, count = 0;
	
	for (n = dispstart; n < num_rules; n++) {
		print_rule(rules + n);
		count ++;
		if (maxprint > 0 && count >= maxprint)
			break;
	}
}

/* main program functions */

int
read_view(int dev)
{
	if (curr_mgr == NULL)
		return (0);

	if (curr_mgr->read_fn != NULL)
		return (curr_mgr->read_fn(dev));

	return (0);
}

void
sort_view(void)
{
	if (curr_mgr != NULL)
		if (curr_mgr->sort_fn != NULL)
			curr_mgr->sort_fn();
}

void
sig_close(int signal)
{
	gotsig_close = 1;
}

void
sig_resize(int signal)
{
	gotsig_resize = 1;
}

void
sig_alarm(int signal)
{
	gotsig_alarm = 1;
}

void
usage()
{
	extern char *__progname;
	fprintf(stderr, "usage: %s [-inbra] [-d count]", __progname);
	fprintf(stderr, " [-s time] [-o field] [number]\n");
	exit(1);
}

void
show_help(void)
{
	int line = 1;
	if (rawmode)
		return;

	erase();
	mvprintw(line, 3, "pfTop Help");
	line += 2;
	mvprintw(line++, 5, " o  - change Ordering");
	mvprintw(line++, 5, " r  - change order diRection");
	mvprintw(line++, 5, " n  - No ordering ");
	mvprintw(line++, 5, " a  - order by Age");
	mvprintw(line++, 5, " e  - order by Expiry time");
	mvprintw(line++, 5, " b  - order by number of Bytes");
	mvprintw(line++, 5, " p  - order by number of packets");
	mvprintw(line++, 5, " s  - order by Source address");
	mvprintw(line++, 5, " S  - order by Source port");
	mvprintw(line++, 5, " d  - order by Destination address");
	mvprintw(line++, 5, " D  - order by Destination port");
	mvprintw(line++, 5, " v  - change View");
	mvprintw(line++, 5, "0-4 - set state view");
	mvprintw(line++, 5, "5,6 - set rule view");
	mvprintw(line++, 5, " h  - Help (this page)");
	mvprintw(line++, 5, " q  - Quit");
	line ++;
	mvprintw(line++, 5, " cursor keys  - scroll display");
	line ++;
	mvprintw(line++, 6, "press any key to continue ...");

	while (getch() == ERR);
}

void
setup_term(int maxstates)
{
	maxprint = maxstates;

	if (rawmode) {
		columns = 80;
		lines = 25;
		clear_linebuf();
	} else {
		screen = newterm(NULL, stdout, stdin);
		if (screen == NULL) {
			rawmode = 1;
			interactive = 0;
			setup_term(maxstates);
			return;
		}
		columns = COLS;
		lines = LINES;

		if (maxstates > lines - HEADER_LINES)
			maxprint = lines - HEADER_LINES;

		nonl();
		keypad (stdscr, TRUE);
		intrflush(stdscr, FALSE);

		halfdelay(10);
		noecho();
	}

	if (maxstates == 0)
		maxprint = lines - HEADER_LINES;

	field_setup();
}

void
keyboard(void)
{
	int ch;

	ch = getch();

	switch (ch) {
	case 'q':
	case 'Q':
		gotsig_close = 1;
		break;
	case 'r':
	case 'R':
		sortdir *= -1;
		need_sort = 1;
		break;
	case 'v':
	case 'V':
		next_view();
		break;
	case 'o':
	case 'O':
		next_order();
		need_sort = 1;
		break;
	case '?':
	case 'h':
	case 'H':
		show_help();
		need_update = 1;
		break;
	case KEY_DOWN:
	case CTRL_N:
		dispstart ++;
		need_update = 1;
		break;
	case KEY_UP:
	case CTRL_P:
		dispstart --;
		need_update = 1;
		break;
	case KEY_NPAGE:
	case CTRL_V:
		dispstart += maxprint;
		need_update = 1;
		break;
	case KEY_PPAGE:
	case META_V:
		dispstart -= maxprint;
		need_update = 1;
		break;
	case KEY_HOME:
	case CTRL_A:
		dispstart = 0;
		need_update = 1;
		break;
	case KEY_END:
	case CTRL_E:
		dispstart = num_disp;
		need_update = 1;
		break;
	default:
		break;
	}

	if (set_order_hotkey(ch))
		need_sort = 1;
	else
		set_view_hotkey(ch);
}

int
main(int argc, char *argv[])
{
	extern char *optarg;
	extern int optind;

	char *orderstr = NULL;
	char *viewstr = NULL;

	int countmax = 0;
	int delay = 5;
	int count = 0;
	int maxstates = 0;

	int dev = -1;
	int ch;

	while ((ch = getopt(argc, argv, "ahnibrs:d:o:v:")) != -1) {
		switch (ch) {
		case 'd':
			countmax = atoi(optarg);
			if (countmax < 0)
				countmax = 0;
			break;
		case 's':
			delay = atoi(optarg);
			if (delay < 1)
				delay = 1;
			break;
		case 'o':
			orderstr = optarg;
			break;
		case 'v':
			viewstr = optarg;
			break;
		case 'i':
			interactive = 1;
			break;
		case 'b':
			rawmode = 1;
		case 'n':
			interactive = 0;
			break;
		case 'r':
			sortdir *= -1;
			break;
		case 'a':
			maxstates = -1;
			break;
		case 'h':
		default:
			usage();
			/* NOTREACHED */
		}
	}

	argc -= optind;
	argv += optind;

	if (argc == 1)
		maxstates = atoi(argv[0]);
	else if (argc > 1)
		usage();

	set_order(orderstr);
	set_view(viewstr);

	if (!isatty(STDOUT_FILENO)) {
		rawmode = 1;
		interactive = 0;
	}
	signal(SIGTERM, sig_close);
	signal(SIGINT, sig_close);
	signal(SIGQUIT, sig_close);
	signal(SIGWINCH, sig_resize);
	signal(SIGALRM, sig_alarm);

#if OS_LEVEL > 30
	dev = open("/dev/pf", O_RDONLY);
#else
	dev = open("/dev/pf", O_RDWR);
#endif
	if (dev == -1)
		err(1, "open(\"/dev/pf\")");

	alloc_buf(0);

	setup_term(maxstates);

	if (rawmode && countmax == 0)
		countmax = 1;

	gotsig_alarm = 1;
	while (1) {
		if (gotsig_alarm) {
			read_view(dev);
			need_sort = 1;
			gotsig_alarm = 0;
			alarm(delay);
		}

		if (need_sort) {
			sort_view();
			need_sort = 0;
			need_update = 1;
		}

		if (need_update) {
			erase();
			disp_update(dev);
			end_page();
			need_update =  0;
		}

		if (countmax && ++count >= countmax)
			break;
		if (gotsig_close)
			break;
		if (gotsig_resize) {
			if (rawmode == 0)
				endwin();
			setup_term(maxstates);
			gotsig_resize = 0;
			need_update = 1;
		}
		if (interactive && need_update == 0)
			keyboard();
	}

	if (rawmode == 0)
		endwin();

	close(dev);
	return 0;
}
